// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_lib_interpolation_vector.hpp"

template <typename T>
void edges2cells_vector_lib(const T *p_vn_in, const T *p_vt_in,
                            const int *cell_edge_idx, const int *cell_edge_blk,
                            const T *e_bln_c_u, const T *e_bln_c_v, T *p_u_out,
                            T *p_v_out,
                            // Additional integer parameters.
                            int i_startblk, int i_endblk, int i_startidx_in,
                            int i_endidx_in, int slev, int elev, int nproma,
                            // Dimensions for the arrays.
                            int nlev, int nblks_e, int nblks_c) {

  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  UnmanagedConstT3D p_vn_in_view(p_vn_in, nproma, nlev, nblks_e);
  UnmanagedConstT3D p_vt_in_view(p_vt_in, nproma, nlev, nblks_e);

  UnmanagedConstInt3D cell_edge_idx_view(cell_edge_idx, nproma, nblks_c, 3);
  UnmanagedConstInt3D cell_edge_blk_view(cell_edge_blk, nproma, nblks_c, 3);

  UnmanagedConstT3D e_bln_c_u_view(e_bln_c_u, nproma, 6, nblks_c);
  UnmanagedConstT3D e_bln_c_v_view(e_bln_c_v, nproma, 6, nblks_c);

  UnmanagedT3D p_u_out_view(p_u_out, nproma, nlev, nblks_c);
  UnmanagedT3D p_v_out_view(p_v_out, nproma, nlev, nblks_c);

  // Loop over cell blocks as in the original Fortran code.
  for (int jb = i_startblk; jb <= i_endblk; ++jb) {
    // Call get_indices_c_lib to get inner loop indices for block jb.
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});
    Kokkos::parallel_for(
        "edges2cells_inner", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          // Compute the bilinear interpolation for cell (jc, jk, jb).
          p_u_out_view(jc, jk, jb) =
              e_bln_c_u_view(jc, 0, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 0) - 1, jk,
                               cell_edge_blk_view(jc, jb, 0) - 1) +
              e_bln_c_u_view(jc, 1, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 0) - 1, jk,
                               cell_edge_blk_view(jc, jb, 0) - 1) +
              e_bln_c_u_view(jc, 2, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 1) - 1, jk,
                               cell_edge_blk_view(jc, jb, 1) - 1) +
              e_bln_c_u_view(jc, 3, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 1) - 1, jk,
                               cell_edge_blk_view(jc, jb, 1) - 1) +
              e_bln_c_u_view(jc, 4, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 2) - 1, jk,
                               cell_edge_blk_view(jc, jb, 2) - 1) +
              e_bln_c_u_view(jc, 5, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 2) - 1, jk,
                               cell_edge_blk_view(jc, jb, 2) - 1);

          p_v_out_view(jc, jk, jb) =
              e_bln_c_v_view(jc, 0, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 0) - 1, jk,
                               cell_edge_blk_view(jc, jb, 0) - 1) +
              e_bln_c_v_view(jc, 1, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 0) - 1, jk,
                               cell_edge_blk_view(jc, jb, 0) - 1) +
              e_bln_c_v_view(jc, 2, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 1) - 1, jk,
                               cell_edge_blk_view(jc, jb, 1) - 1) +
              e_bln_c_v_view(jc, 3, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 1) - 1, jk,
                               cell_edge_blk_view(jc, jb, 1) - 1) +
              e_bln_c_v_view(jc, 4, jb) *
                  p_vn_in_view(cell_edge_idx_view(jc, jb, 2) - 1, jk,
                               cell_edge_blk_view(jc, jb, 2) - 1) +
              e_bln_c_v_view(jc, 5, jb) *
                  p_vt_in_view(cell_edge_idx_view(jc, jb, 2) - 1, jk,
                               cell_edge_blk_view(jc, jb, 2) - 1);
        });
    // Optionally fence after each block if required.
    Kokkos::fence();
  }
}

template void edges2cells_vector_lib<double>(
    const double *p_vn_in, const double *p_vt_in, const int *cell_edge_idx,
    const int *cell_edge_blk, const double *e_bln_c_u, const double *e_bln_c_v,
    double *p_u_out, double *p_v_out,
    // Additional integer parameters.
    int i_startblk, int i_endblk, int i_startidx_in, int i_endidx_in, int slev,
    int elev, int nproma,
    // Dimensions for the arrays.
    int nlev, int nblks_e, int nblks_c);

template void edges2cells_vector_lib<float>(
    const float *p_vn_in, const float *p_vt_in, const int *cell_edge_idx,
    const int *cell_edge_blk, const float *e_bln_c_u, const float *e_bln_c_v,
    float *p_u_out, float *p_v_out,
    // Additional integer parameters.
    int i_startblk, int i_endblk, int i_startidx_in, int i_endidx_in, int slev,
    int elev, int nproma,
    // Dimensions for the arrays.
    int nlev, int nblks_e, int nblks_c);