// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#pragma once

// Template function for computing array size.
// For example, we get the array size of a 4-dimensional array A(2, 3, 4, 5) by
//    dim_combine(2, 3, 4, 5).
// Which will automatically instantiate
//    dim_combine<int, int, int, int>(2, 3, 4, 5).
// The function then call dim_combine recursively
//    dim_combine<int, int, int, int>(2, 3, 4, 5) {
//      return static_cast<size_t>(2) * dim_combine<int, int, int>(3, 4, 5);
//    }
//    dim_combine<int, int, int>(3, 4, 5) {
//      return static_cast<size_t>(3) * dim_combine<int, int>(4, 5);
//    }
//    dim_combine<int, int>(4, 5) {
//      return static_cast<size_t>(4) * dim_combine<int>(5);
//    }
// Where the last dim_combine is specialized as
//    dim_combine<int>(5) {
//      return static_cast<size_t>(5);
//    }
// Which gives
//    dim_combine<int, int, int, int>(2, 3, 4, 5) =
//      static_cast<size_t>(2) * static_cast<size_t>(3) *
//      static_cast<size_t>(4) * static_cast<size_t>(5)
/// Template helpers for combining multiple dimension array sizes.
/// The base function of dimension combine. Should not be used.
template <typename... Ts> size_t dim_combine(Ts... dims) { return 0; }
/// Template specialization of only one dimension, returns the dimension itself.
template <typename T> size_t dim_combine(T dim) {
  return static_cast<size_t>(dim);
}
/// Template specialization of picking out the first dimension. The combined
/// dimension is the first dimension times the combined dimension of the rest.
template <typename T, typename... Ts> size_t dim_combine(T dim, Ts... dims) {
  return static_cast<size_t>(dim) * dim_combine(dims...);
}

// Template function for LayoutLeft ID access in compile time.
// For example, a multi-dimensional array A of dimensions <2, 3, 4, 5> gets its
// corresponding vector id (LayoutLeft) by
//    at<2, 3, 4, 5>(id1, id2, id3, id4).
// The at_impl then adds the id from beginning to the end and pass the id prefix
// to the next recursive at_impl function. In this example,
//    at<2, 3, 4, 5>(id1, id2, id3, id4) {
//      return id1 + at_impl<3, 4, 5>(2, id2, id3, id4);
//    }
//    at_impl<3, 4, 5>(2, id2, id3, id4) {
//      return id2 * 2 + at_impl<4, 5>(2 * 3, id3, id4);
//    }
//    at_impl<4, 5>(2 * 3, id3, id4) {
//      return id3 * 2 * 3 + at_impl<5>(2 * 3 * 4, id4);
//    }
//    at_impl<5>(2 * 3 * 4, id4) {
//      return id4 * 2 * 3 * 4;
//    }
// Which gives
//    at<2, 3, 4, 5>(id1, id2, id3, id4) = id1         + id2 * 2         +
//                                         id3 * 2 * 3 + id4 * 2 * 3 * 4
/// Helper type converting integer numbers to int
template <class T, auto> using always_t = T;
/// Base function of at_impl. Should not be used.
template <int... Dims> int at_impl(always_t<int, Dims>... ids) { return 0; }
/// Template specialization of the last ID
template <int LastDim> int at_impl(int prefix, int id) { return id * prefix; }
/// Template specialization of at_impl, accumulate the return value using the
/// first id and pass the prefix to the next recursive at_impl function.
template <int FirstDim, int... Dims>
int at_impl(int prefix, int id, always_t<int, Dims>... ids) {
  return id * prefix + at_impl<Dims...>(prefix * FirstDim, ids...);
}
/// at<dim1, dim2, ...>(id1, id2, ...) gets its memory index in vector assuming
/// LayoutLeft. Use this function instead of at_impl.
template <int FirstDim, int... Dims>
int at(int id, always_t<int, Dims>... ids) {
  return id + at_impl<Dims...>(FirstDim, ids...);
}
