// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include <iostream>
#include <random>
#include <vector>

#include <Kokkos_Core.hpp>
#include <gtest/gtest.h>
#include <dim_helper.hpp>
#include <horizontal/mo_lib_divrot.hpp>
#include <support/mo_lib_loopindices.hpp>

/// Test class for the horizontal divergence tests. Templated for the ValueType
template <typename ValueType> class HorizontalDivTest : public ::testing::Test {
protected:
  static constexpr int nproma = 3;  // inner loop length
  static constexpr int nlev = 2;    // number of vertical levels
  static constexpr int nblks_c = 1; // number of cell blocks
  static constexpr int nblks_e = 1; // number of edge blocks
  static constexpr int dim4d = 2;   // 4th dimension size

  int i_startblk = 0;
  int i_endblk = nblks_c; // Test blocks [0 .. nblks_c-1]
  int i_startidx_in = 0;
  int i_endidx_in = nproma; // Full range: 0 .. nproma-1
  std::vector<int> slev;
  std::vector<int> elev;
  bool lacc = false; // Not using ACC-specific behavior.

  std::vector<ValueType> vec_e;
  std::vector<int> cell_edge_idx;
  std::vector<int> cell_edge_blk;
  std::vector<ValueType> geofac_div;
  std::vector<ValueType> div_vec_c;
  std::vector<ValueType> f4din;
  std::vector<ValueType> f4dout;

  // Followings are needed in HorizontalDivAvgTest
  std::vector<int> cell_neighbor_idx;
  std::vector<int> cell_neighbor_blk;
  std::vector<ValueType> avg_coeff;
  std::vector<ValueType> opt_in2;
  std::vector<ValueType> opt_out2;

  HorizontalDivTest() {
    slev.resize(dim4d, 0);
    elev.resize(dim4d, nlev); // Full vertical range (0 .. nlev-1)

    vec_e.resize(dim_combine(nproma, nlev, nblks_e));
    cell_edge_idx.resize(dim_combine(nproma, nblks_c, 3));
    cell_edge_blk.resize(dim_combine(nproma, nblks_c, 3));
    geofac_div.resize(dim_combine(nproma, 3, nblks_c));
    div_vec_c.resize(dim_combine(nproma, nlev, nblks_c));
    f4din.resize(dim_combine(nproma, nlev, nblks_e, dim4d));
    f4dout.resize(dim_combine(nproma, nlev, nblks_c, dim4d));
    cell_neighbor_idx.resize(dim_combine(nproma, nblks_c, 3));
    cell_neighbor_blk.resize(dim_combine(nproma, nblks_c, 3));
    avg_coeff.resize(dim_combine(nproma, 4, nblks_c));
    opt_in2.resize(dim_combine(nproma, nlev, nblks_e));
    opt_out2.resize(dim_combine(nproma, nlev, nblks_c));
  }
};

/// ValueTypes which the divrot tests should run with
typedef ::testing::Types<float, double> ValueTypes;

TYPED_TEST_SUITE(HorizontalDivTest, ValueTypes);

TYPED_TEST(HorizontalDivTest, TestDiv3DSpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Initialization with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
    }

    // Set edge indices to point to specific cells (including self)
    this->cell_edge_idx[cell_edge_at(i, 0, 0)] = i;
    this->cell_edge_idx[cell_edge_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_edge_idx[cell_edge_at(i, 0, 2)] = (i + 2) % nproma;

    // All edges are in the same block for this test
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
    }

    // Geometric factors
    this->geofac_div[geofac_div_at(i, 0, 0)] = 0.5;
    this->geofac_div[geofac_div_at(i, 1, 0)] = 0.3;
    this->geofac_div[geofac_div_at(i, 2, 0)] = 0.2;

    // Initialize div_vec_c to zero
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = 0.0;
    }
  }

  // Call the div3d function
  div3d<TypeParam>(this->vec_e.data(), this->cell_edge_idx.data(),
                   this->cell_edge_blk.data(), this->geofac_div.data(),
                   this->div_vec_c.data(), this->i_startblk, this->i_endblk,
                   this->i_startidx_in, this->i_endidx_in, this->slev[0],
                   this->elev[0], this->nproma, this->lacc, this->nlev,
                   this->nblks_c, this->nblks_e);

  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 1, 0)], 3.4, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 0, 0)], 2.1, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 1, 0)], 4.2, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 0, 0)], 2.2, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 1, 0)], 4.4, 1e-6);
}

TYPED_TEST(HorizontalDivTest, TestDiv3DRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialization with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
    }

    // Set random edge indices
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = int_distrib(gen);
      this->cell_edge_blk[cell_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 3; ++j) {
      this->geofac_div[geofac_div_at(i, j, 0)] = real_distrib(gen);
    }

    // Initialize div_vec_c to random values
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = real_distrib(gen);
    }
  }

  // Call the div3d function
  div3d<TypeParam>(this->vec_e.data(), this->cell_edge_idx.data(),
                   this->cell_edge_blk.data(), this->geofac_div.data(),
                   this->div_vec_c.data(), this->i_startblk, this->i_endblk,
                   this->i_startidx_in, this->i_endidx_in, this->slev[0],
                   this->elev[0], this->nproma, this->lacc, this->nlev,
                   this->nblks_c, this->nblks_e);

  // Calculate reference values separately and verify results
  std::vector<TypeParam> ref_div_vec_c(nproma * nlev * nblks_c, 0.0);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 0)])] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 1)])] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 2)])] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];
      }
    }
  }

  // Verify results
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->div_vec_c[div_vec_c_at(i, k, 0)],
                  ref_div_vec_c[div_vec_c_at(i, k, 0)], 1e-5)
          << "Results differ at i=" << i << ", k=" << k;
    }
  }
}

TYPED_TEST(HorizontalDivTest, TestDiv3D2FSpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;
  const auto &f4d_at = at<nproma, nlev, nblks_e, dim4d>;
  const auto &f4dout_at = at<nproma, nlev, nblks_c, dim4d>;

  // Initialization with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
      this->f4din[f4d_at(i, k, 0, 0)] =
          (i + 1) * (k + 2); // Different pattern for second field
    }

    // Set edge indices to point to specific cells (including self)
    this->cell_edge_idx[cell_edge_at(i, 0, 0)] = i;
    this->cell_edge_idx[cell_edge_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_edge_idx[cell_edge_at(i, 0, 2)] = (i + 2) % nproma;

    // All edges are in the same block for this test
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
    }

    // Geometric factors
    this->geofac_div[geofac_div_at(i, 0, 0)] = 0.5;
    this->geofac_div[geofac_div_at(i, 1, 0)] = 0.3;
    this->geofac_div[geofac_div_at(i, 2, 0)] = 0.2;

    // Initialize div_vec_c and f4dout to zero
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = 0.0;
      this->f4dout[f4dout_at(i, k, 0, 0)] = 0.0;
    }
  }

  // Call the div3d_2field function
  div3d_2field<TypeParam>(this->vec_e.data(), this->cell_edge_idx.data(),
                          this->cell_edge_blk.data(), this->geofac_div.data(),
                          this->div_vec_c.data(), this->f4din.data(),
                          this->f4dout.data(), this->i_startblk, this->i_endblk,
                          this->i_startidx_in, this->i_endidx_in, this->slev[0],
                          this->elev[0], this->nproma, this->lacc, this->nlev,
                          this->nblks_c, this->nblks_e);

  // Check first field (same as in div3d test)
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 1, 0)], 3.4, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 0, 0)], 2.1, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 1, 0)], 4.2, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 0, 0)], 2.2, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 1, 0)], 4.4, 1e-6);

  // Check second field (expected values calculated manually)
  EXPECT_NEAR(this->f4dout[f4dout_at(0, 0, 0, 0)], 3.4, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(0, 1, 0, 0)], 5.1, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 0, 0, 0)], 4.2, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 1, 0, 0)], 6.3, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 0, 0, 0)], 4.4, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 1, 0, 0)], 6.6, 1e-6);
}

TYPED_TEST(HorizontalDivTest, TestDiv3D2FRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;
  const auto &f4d_at = at<nproma, nlev, nblks_e, dim4d>;
  const auto &f4dout_at = at<nproma, nlev, nblks_c, dim4d>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialization with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
      this->f4din[f4d_at(i, k, 0, 0)] = real_distrib(gen);
    }

    // Set random edge indices
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = int_distrib(gen);
      this->cell_edge_blk[cell_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 3; ++j) {
      this->geofac_div[geofac_div_at(i, j, 0)] = real_distrib(gen);
    }

    // Initialize div_vec_c and f4dout to random values
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = real_distrib(gen);
      this->f4dout[f4dout_at(i, k, 0, 0)] = real_distrib(gen);
    }
  }

  // Call the div3d_2field function
  div3d_2field<TypeParam>(this->vec_e.data(), this->cell_edge_idx.data(),
                          this->cell_edge_blk.data(), this->geofac_div.data(),
                          this->div_vec_c.data(), this->f4din.data(),
                          this->f4dout.data(), this->i_startblk, this->i_endblk,
                          this->i_startidx_in, this->i_endidx_in, this->slev[0],
                          this->elev[0], this->nproma, this->lacc, this->nlev,
                          this->nblks_c, this->nblks_e);

  // Calculate reference values separately and verify results
  std::vector<TypeParam> ref_div_vec_c(nproma * nlev * nblks_c, 0.0);
  std::vector<TypeParam> ref_f4dout(nproma * nlev * nblks_c * dim4d, 0.0);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        // Calculate reference value for first field
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 0)])] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 1)])] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 2)])] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];

        // Calculate reference value for second field
        ref_f4dout[f4dout_at(jc, jk, jb, 0)] =
            this->f4din[f4d_at(this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                               this->cell_edge_blk[cell_edge_at(jc, jb, 0)],
                               0)] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->f4din[f4d_at(this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                               this->cell_edge_blk[cell_edge_at(jc, jb, 1)],
                               0)] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->f4din[f4d_at(this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                               this->cell_edge_blk[cell_edge_at(jc, jb, 2)],
                               0)] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];
      }
    }
  }

  // Verify results for first field
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->div_vec_c[div_vec_c_at(i, k, 0)],
                  ref_div_vec_c[div_vec_c_at(i, k, 0)], 1e-5)
          << "First field results differ at i=" << i << ", k=" << k;
    }
  }

  // Verify results for second field
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->f4dout[f4dout_at(i, k, 0, 0)],
                  ref_f4dout[f4dout_at(i, k, 0, 0)], 1e-5)
          << "Second field results differ at i=" << i << ", k=" << k;
    }
  }
}

TYPED_TEST(HorizontalDivTest, TestDiv4DSpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &f4din_at = at<nproma, nlev, nblks_e, dim4d>;
  const auto &f4dout_at = at<nproma, nlev, nblks_c, dim4d>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = (i + j) % nproma;
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
      this->geofac_div[geofac_div_at(i, j, 0)] = 0.1 * (j + 1);
    }

    for (int k = 0; k < nlev; ++k) {
      for (int d = 0; d < dim4d; ++d) {
        this->f4din[f4din_at(i, k, 0, d)] = 1.0 + i + k + d;
        this->f4dout[f4dout_at(i, k, 0, d)] = 0.0;
      }
    }
  }

  // Test function
  div4d<TypeParam>(this->cell_edge_idx.data(), this->cell_edge_blk.data(),
                   this->geofac_div.data(), this->f4din.data(),
                   this->f4dout.data(), this->dim4d, this->i_startblk,
                   this->i_endblk, this->i_startidx_in, this->i_endidx_in,
                   this->slev.data(), this->elev.data(), this->nproma,
                   this->lacc, this->nlev, this->nblks_c, this->nblks_e);

  EXPECT_NEAR(this->f4dout[f4dout_at(0, 0, 0, 0)], 1.4, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 0, 0, 0)], 1.1, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 0, 0, 0)], 1.1, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(0, 1, 0, 0)], 2.0, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 1, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 1, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(0, 0, 0, 1)], 2.0, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 0, 0, 1)], 1.7, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 0, 0, 1)], 1.7, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(0, 1, 0, 1)], 2.6, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(1, 1, 0, 1)], 2.3, 1e-6);
  EXPECT_NEAR(this->f4dout[f4dout_at(2, 1, 0, 1)], 2.3, 1e-6);
}

TYPED_TEST(HorizontalDivTest, TestDiv4DRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &f4din_at = at<nproma, nlev, nblks_e, dim4d>;
  const auto &f4dout_at = at<nproma, nlev, nblks_c, dim4d>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 3.0);

  // Initialize with random values
  for (int i = 0; i < nproma; ++i) {
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = int_distrib(gen);
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
      this->geofac_div[geofac_div_at(i, j, 0)] = real_distrib(gen);
    }

    for (int k = 0; k < nlev; ++k) {
      for (int d = 0; d < dim4d; ++d) {
        this->f4din[f4din_at(i, k, 0, d)] = real_distrib(gen);
        this->f4dout[f4dout_at(i, k, 0, d)] = 0.0;
      }
    }
  }

  // Test function
  div4d<TypeParam>(this->cell_edge_idx.data(), this->cell_edge_blk.data(),
                   this->geofac_div.data(), this->f4din.data(),
                   this->f4dout.data(), this->dim4d, this->i_startblk,
                   this->i_endblk, this->i_startidx_in, this->i_endidx_in,
                   this->slev.data(), this->elev.data(), this->nproma,
                   this->lacc, this->nlev, this->nblks_c, this->nblks_e);

  // Compute reference result and check
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);

    for (int ji = 0; ji < dim4d; ++ji) {
      for (int jk = this->slev[ji]; jk < this->elev[ji]; ++jk) {
        for (int jc = i_startidx; jc < i_endidx; ++jc) {
          TypeParam expected = 0.0;
          for (int je = 0; je < 3; ++je) {
            expected +=
                this->f4din[f4din_at(
                    this->cell_edge_idx[cell_edge_at(jc, jb, je)], jk,
                    this->cell_edge_blk[cell_edge_at(jc, jb, je)], ji)] *
                this->geofac_div[geofac_div_at(jc, je, jb)];
          }

          EXPECT_NEAR(this->f4dout[f4dout_at(jc, jk, jb, ji)], expected, 1e-5)
              << "Random test fails at jc=" << jc << ", jk=" << jk
              << ", jb=" << jb << ", ji=" << ji;
        }
      }
    }
  }
}

TYPED_TEST_SUITE(HorizontalDivTest, ValueTypes);

TYPED_TEST(HorizontalDivTest, TestDivAvgSpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Vectors for additional parameters
  // Vectors for block and index ranges
  std::vector<int> i_startblk_in(3, 0);
  std::vector<int> i_endblk_in(3, nblks_c);
  std::vector<int> i_startidx_in(3, 0);
  std::vector<int> i_endidx_in(3, nproma);

  // Parameters for the test
  int patch_id = 1;
  bool l_limited_area = true;
  bool l2fields = true;

  const auto &cell_neighbor_at = at<nproma, nblks_c, 3>;
  const auto &avg_coeff_at = at<nproma, 4, nblks_c>;

  // Initialize the vectors with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
      this->opt_in2[vec_e_at(i, k, 0)] =
          (i + 1) * (k + 1) * 0.5; // Half of vec_e
    }

    // Set edge indices to point to specific cells
    this->cell_edge_idx[cell_edge_at(i, 0, 0)] = i;
    this->cell_edge_idx[cell_edge_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_edge_idx[cell_edge_at(i, 0, 2)] = (i + 2) % nproma;

    // Set neighbor indices similarly
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = i;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 2)] = (i + 2) % nproma;

    // All edges and neighbors are in the same block for this test
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    // Geometric factors
    this->geofac_div[geofac_div_at(i, 0, 0)] = 0.5;
    this->geofac_div[geofac_div_at(i, 1, 0)] = 0.3;
    this->geofac_div[geofac_div_at(i, 2, 0)] = 0.2;

    // Average coefficients
    this->avg_coeff[avg_coeff_at(i, 0, 0)] = 0.4; // Self
    this->avg_coeff[avg_coeff_at(i, 1, 0)] = 0.2; // First neighbor
    this->avg_coeff[avg_coeff_at(i, 2, 0)] = 0.2; // Second neighbor
    this->avg_coeff[avg_coeff_at(i, 3, 0)] = 0.2; // Third neighbor

    // Initialize div_vec_c and opt_out2 to zero
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = 0.0;
      this->opt_out2[div_vec_c_at(i, k, 0)] = 0.0;
    }
  }

  // Call the div_avg function
  div_avg<TypeParam>(
      this->vec_e.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->cell_edge_idx.data(),
      this->cell_edge_blk.data(), this->geofac_div.data(),
      this->avg_coeff.data(), this->div_vec_c.data(), this->opt_in2.data(),
      this->opt_out2.data(), i_startblk_in.data(), i_endblk_in.data(),
      i_startidx_in.data(), i_endidx_in.data(), this->slev[0], this->elev[0],
      this->nproma, patch_id, l_limited_area, l2fields, this->lacc, this->nlev,
      this->nblks_c, this->nblks_e);

  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 0, 0)], 1.88, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 1, 0)], 3.76, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 0, 0)], 2.04, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 1, 0)], 4.08, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 0, 0)], 2.08, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 1, 0)], 4.16, 1e-6);

  EXPECT_NEAR(this->opt_out2[div_vec_c_at(0, 0, 0)], 0.94, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(0, 1, 0)], 1.88, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(1, 0, 0)], 1.02, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(1, 1, 0)], 2.04, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(2, 0, 0)], 1.04, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(2, 1, 0)], 2.08, 1e-6);
}

TYPED_TEST(HorizontalDivTest, TestDivAvgRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Vectors for block and index ranges
  std::vector<int> i_startblk_in(3, 0);
  std::vector<int> i_endblk_in(3, nblks_c);
  std::vector<int> i_startidx_in(3, 0);
  std::vector<int> i_endidx_in(3, nproma);

  // Parameters for the test
  int patch_id = 1;
  bool l_limited_area = true;
  bool l2fields = true;

  const auto &cell_neighbor_at = at<nproma, nblks_c, 3>;
  const auto &avg_coeff_at = at<nproma, 4, nblks_c>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialize with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
      this->opt_in2[vec_e_at(i, k, 0)] = real_distrib(gen);
    }

    // Set random edge indices
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = int_distrib(gen);
      this->cell_edge_blk[cell_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity

      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 3; ++j) {
      this->geofac_div[geofac_div_at(i, j, 0)] = real_distrib(gen);
    }

    // Random average coefficients
    for (int j = 0; j < 4; ++j) {
      this->avg_coeff[avg_coeff_at(i, j, 0)] = real_distrib(gen);
    }

    // Random initial values for div_vec_c and opt_out2
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = real_distrib(gen);
      this->opt_out2[div_vec_c_at(i, k, 0)] = real_distrib(gen);
    }
  }

  // Call the div_avg function
  div_avg<TypeParam>(
      this->vec_e.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->cell_edge_idx.data(),
      this->cell_edge_blk.data(), this->geofac_div.data(),
      this->avg_coeff.data(), this->div_vec_c.data(), this->opt_in2.data(),
      this->opt_out2.data(), i_startblk_in.data(), i_endblk_in.data(),
      i_startidx_in.data(), i_endidx_in.data(), this->slev[0], this->elev[0],
      this->nproma, patch_id, l_limited_area, l2fields, this->lacc, this->nlev,
      this->nblks_c, this->nblks_e);

  // Calculate reference values manually
  std::vector<TypeParam> aux_c(dim_combine(nproma, nlev, nblks_c));
  std::vector<TypeParam> aux_c2(dim_combine(nproma, nlev, nblks_c));
  std::vector<TypeParam> ref_div_vec_c(dim_combine(nproma, nlev, nblks_c));
  std::vector<TypeParam> ref_opt_out2(dim_combine(nproma, nlev, nblks_c));

  // Step 1: Calculate aux_c and aux_c2
  for (int jb = i_startblk_in[0]; jb < i_endblk_in[0]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[0], i_endidx_in[0], nproma, jb,
                      i_startblk_in[0], i_endblk_in[0], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        aux_c[div_vec_c_at(jc, jk, jb)] =
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 0)])] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 1)])] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 2)])] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];

        aux_c2[div_vec_c_at(jc, jk, jb)] =
            this->opt_in2[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 0)])] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->opt_in2[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 1)])] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->opt_in2[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 2)])] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];
      }
    }
  }

  // Step 2: Assign aux_c to div_vec_c and aux_c2 to opt_out2 for patch_id > 0
  for (int jb = i_startblk_in[1]; jb < i_endblk_in[1]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[1], i_endidx_in[1], nproma, jb,
                      i_startblk_in[1], i_endblk_in[1], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            aux_c[div_vec_c_at(jc, jk, jb)];
        ref_opt_out2[div_vec_c_at(jc, jk, jb)] =
            aux_c2[div_vec_c_at(jc, jk, jb)];
      }
    }
  }

  // Step 3: Perform averaging for the rest of the blocks
  for (int jb = i_startblk_in[2]; jb < i_endblk_in[2]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[2], i_endidx_in[2], nproma, jb,
                      i_startblk_in[2], i_endblk_in[2], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            aux_c[div_vec_c_at(jc, jk, jb)] *
                this->avg_coeff[avg_coeff_at(jc, 0, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 0)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 0)])] *
                this->avg_coeff[avg_coeff_at(jc, 1, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 1)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 1)])] *
                this->avg_coeff[avg_coeff_at(jc, 2, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 2)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 2)])] *
                this->avg_coeff[avg_coeff_at(jc, 3, jb)];

        ref_opt_out2[div_vec_c_at(jc, jk, jb)] =
            aux_c2[div_vec_c_at(jc, jk, jb)] *
                this->avg_coeff[avg_coeff_at(jc, 0, jb)] +
            aux_c2[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 0)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 0)])] *
                this->avg_coeff[avg_coeff_at(jc, 1, jb)] +
            aux_c2[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 1)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 1)])] *
                this->avg_coeff[avg_coeff_at(jc, 2, jb)] +
            aux_c2[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 2)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 2)])] *
                this->avg_coeff[avg_coeff_at(jc, 3, jb)];
      }
    }
  }

  // Verify results
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->div_vec_c[div_vec_c_at(i, k, 0)],
                  ref_div_vec_c[div_vec_c_at(i, k, 0)], 1e-5)
          << "div_vec_c results differ at i=" << i << ", k=" << k;

      EXPECT_NEAR(this->opt_out2[div_vec_c_at(i, k, 0)],
                  ref_opt_out2[div_vec_c_at(i, k, 0)], 1e-5)
          << "opt_out2 results differ at i=" << i << ", k=" << k;
    }
  }
}

TYPED_TEST(HorizontalDivTest, TestDivAvgSpecificNoL2fields) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;
  constexpr int dim4d = this->dim4d;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Vectors for block and index ranges
  std::vector<int> i_startblk_in(3, 0);
  std::vector<int> i_endblk_in(3, nblks_c);
  std::vector<int> i_startidx_in(3, 0);
  std::vector<int> i_endidx_in(3, nproma);

  // Parameters for the test
  int patch_id = 1;
  bool l_limited_area = true;
  bool l2fields = false;

  const auto &cell_neighbor_at = at<nproma, nblks_c, 3>;
  const auto &avg_coeff_at = at<nproma, 4, nblks_c>;

  // Initialize the vectors with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
      this->opt_in2[vec_e_at(i, k, 0)] =
          (i + 1) * (k + 1) * 0.5; // Half of vec_e
    }

    // Set edge indices to point to specific cells
    this->cell_edge_idx[cell_edge_at(i, 0, 0)] = i;
    this->cell_edge_idx[cell_edge_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_edge_idx[cell_edge_at(i, 0, 2)] = (i + 2) % nproma;

    // Set neighbor indices similarly
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = i;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 1)] = (i + 1) % nproma;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 2)] = (i + 2) % nproma;

    // All edges and neighbors are in the same block for this test
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_blk[cell_edge_at(i, 0, j)] = 0;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    // Geometric factors
    this->geofac_div[geofac_div_at(i, 0, 0)] = 0.5;
    this->geofac_div[geofac_div_at(i, 1, 0)] = 0.3;
    this->geofac_div[geofac_div_at(i, 2, 0)] = 0.2;

    // Average coefficients
    this->avg_coeff[avg_coeff_at(i, 0, 0)] = 0.4; // Self
    this->avg_coeff[avg_coeff_at(i, 1, 0)] = 0.2; // First neighbor
    this->avg_coeff[avg_coeff_at(i, 2, 0)] = 0.2; // Second neighbor
    this->avg_coeff[avg_coeff_at(i, 3, 0)] = 0.2; // Third neighbor

    // Initialize div_vec_c and opt_out2 to zero
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = 0.0;
      this->opt_out2[div_vec_c_at(i, k, 0)] = 0.0;
    }
  }

  // Call the div_avg function
  div_avg<TypeParam>(
      this->vec_e.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->cell_edge_idx.data(),
      this->cell_edge_blk.data(), this->geofac_div.data(),
      this->avg_coeff.data(), this->div_vec_c.data(), this->opt_in2.data(),
      this->opt_out2.data(), i_startblk_in.data(), i_endblk_in.data(),
      i_startidx_in.data(), i_endidx_in.data(), this->slev[0], this->elev[0],
      this->nproma, patch_id, l_limited_area, l2fields, this->lacc, this->nlev,
      this->nblks_c, this->nblks_e);

  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 0, 0)], 1.88, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(0, 1, 0)], 3.76, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 0, 0)], 2.04, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(1, 1, 0)], 4.08, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 0, 0)], 2.08, 1e-6);
  EXPECT_NEAR(this->div_vec_c[div_vec_c_at(2, 1, 0)], 4.16, 1e-6);

  EXPECT_NEAR(this->opt_out2[div_vec_c_at(0, 0, 0)], 0.0, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(0, 1, 0)], 0.0, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(1, 0, 0)], 0.0, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(1, 1, 0)], 0.0, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(2, 0, 0)], 0.0, 1e-6);
  EXPECT_NEAR(this->opt_out2[div_vec_c_at(2, 1, 0)], 0.0, 1e-6);
}

TYPED_TEST(HorizontalDivTest, TestDivAvgRandomNoL2fields) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int nblks_e = this->nblks_e;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &cell_edge_at = at<nproma, nblks_c, 3>;
  const auto &geofac_div_at = at<nproma, 3, nblks_c>;
  const auto &div_vec_c_at = at<nproma, nlev, nblks_c>;

  // Vectors for block and index ranges
  std::vector<int> i_startblk_in(3, 0);
  std::vector<int> i_endblk_in(3, nblks_c);
  std::vector<int> i_startidx_in(3, 0);
  std::vector<int> i_endidx_in(3, nproma);

  // Parameters for the test
  int patch_id = 1;
  bool l_limited_area = true;
  bool l2fields = false; // Set to false for this test

  const auto &cell_neighbor_at = at<nproma, nblks_c, 3>;
  const auto &avg_coeff_at = at<nproma, 4, nblks_c>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialize with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
      this->opt_in2[vec_e_at(i, k, 0)] =
          real_distrib(gen); // Not used but initialize anyway
    }

    // Set random edge indices
    for (int j = 0; j < 3; ++j) {
      this->cell_edge_idx[cell_edge_at(i, 0, j)] = int_distrib(gen);
      this->cell_edge_blk[cell_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity

      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 3; ++j) {
      this->geofac_div[geofac_div_at(i, j, 0)] = real_distrib(gen);
    }

    // Random average coefficients
    for (int j = 0; j < 4; ++j) {
      this->avg_coeff[avg_coeff_at(i, j, 0)] = real_distrib(gen);
    }

    // Random initial values for div_vec_c and opt_out2
    for (int k = 0; k < nlev; ++k) {
      this->div_vec_c[div_vec_c_at(i, k, 0)] = real_distrib(gen);
      this->opt_out2[div_vec_c_at(i, k, 0)] =
          real_distrib(gen); // Not used but initialize anyway
    }
  }

  // Call the div_avg function with l2fields=false
  div_avg<TypeParam>(
      this->vec_e.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->cell_edge_idx.data(),
      this->cell_edge_blk.data(), this->geofac_div.data(),
      this->avg_coeff.data(), this->div_vec_c.data(), this->opt_in2.data(),
      this->opt_out2.data(), i_startblk_in.data(), i_endblk_in.data(),
      i_startidx_in.data(), i_endidx_in.data(), this->slev[0], this->elev[0],
      this->nproma, patch_id, l_limited_area, l2fields, this->lacc, this->nlev,
      this->nblks_c, this->nblks_e);

  // Calculate reference values manually
  std::vector<TypeParam> aux_c(dim_combine(nproma, nlev, nblks_c));
  std::vector<TypeParam> ref_div_vec_c(dim_combine(nproma, nlev, nblks_c));

  // Step 1: Calculate aux_c (but not aux_c2 since l2fields=false)
  for (int jb = i_startblk_in[0]; jb < i_endblk_in[0]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[0], i_endidx_in[0], nproma, jb,
                      i_startblk_in[0], i_endblk_in[0], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        aux_c[div_vec_c_at(jc, jk, jb)] =
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 0)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 0)])] *
                this->geofac_div[geofac_div_at(jc, 0, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 1)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 1)])] *
                this->geofac_div[geofac_div_at(jc, 1, jb)] +
            this->vec_e[vec_e_at(
                this->cell_edge_idx[cell_edge_at(jc, jb, 2)], jk,
                this->cell_edge_blk[cell_edge_at(jc, jb, 2)])] *
                this->geofac_div[geofac_div_at(jc, 2, jb)];
      }
    }
  }

  // Step 2: Assign aux_c to div_vec_c for patch_id > 0 (opt_out2 not updated
  // since l2fields=false)
  for (int jb = i_startblk_in[1]; jb < i_endblk_in[1]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[1], i_endidx_in[1], nproma, jb,
                      i_startblk_in[1], i_endblk_in[1], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            aux_c[div_vec_c_at(jc, jk, jb)];
      }
    }
  }

  // Step 3: Perform averaging for the rest of the blocks (only for div_vec_c,
  // not opt_out2)
  for (int jb = i_startblk_in[2]; jb < i_endblk_in[2]; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in[2], i_endidx_in[2], nproma, jb,
                      i_startblk_in[2], i_endblk_in[2], i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        ref_div_vec_c[div_vec_c_at(jc, jk, jb)] =
            aux_c[div_vec_c_at(jc, jk, jb)] *
                this->avg_coeff[avg_coeff_at(jc, 0, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 0)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 0)])] *
                this->avg_coeff[avg_coeff_at(jc, 1, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 1)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 1)])] *
                this->avg_coeff[avg_coeff_at(jc, 2, jb)] +
            aux_c[div_vec_c_at(
                this->cell_neighbor_idx[cell_neighbor_at(jc, jb, 2)], jk,
                this->cell_neighbor_blk[cell_neighbor_at(jc, jb, 2)])] *
                this->avg_coeff[avg_coeff_at(jc, 3, jb)];
      }
    }
  }

  // Verify results - only check div_vec_c since l2fields=false means opt_out2
  // isn't updated
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->div_vec_c[div_vec_c_at(i, k, 0)],
                  ref_div_vec_c[div_vec_c_at(i, k, 0)], 1e-5)
          << "div_vec_c results differ at i=" << i << ", k=" << k;
    }
  }
}
