// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include <iostream>
#include <random>
#include <vector>

#include <Kokkos_Core.hpp>
#include <gtest/gtest.h>
#include <dim_helper.hpp>
#include <horizontal/mo_lib_divrot.hpp>
#include <support/mo_lib_loopindices.hpp>

/// Enum class for the reconstruction method
enum class ReconstructionMethod {
  linear,
  quadratic,
  cubic,
};

/// Base test class for the horizontal reconstruct tests. Templated for the ValueType
/// and ReconMethod for the reconstruction method.
template <typename ValueType, int ReconMethod>
class HorizontalReconTest : public ::testing::Test {
protected:
  // lsq_dim_c and lsq_dim_unk are instantiated in compile time.
  static constexpr std::tuple<int, int>
  init_lsq_dim(ReconstructionMethod method) {
    switch (method) {
    case ReconstructionMethod::linear:
      return std::make_tuple(3, 2);
    case ReconstructionMethod::quadratic:
      return std::make_tuple(9, 5);
    case ReconstructionMethod::cubic:
      return std::make_tuple(9, 9);
    }
  }

  // Constant dimensions.
  static constexpr int nproma = 3;  // inner loop length
  static constexpr int nlev = 1;    // number of vertical levels
  static constexpr int nblks_c = 1; // number of cell blocks (for p_e_in)
  static constexpr std::tuple<int, int> lsq_dim =
      init_lsq_dim(static_cast<ReconstructionMethod>(ReconMethod));
  static constexpr int lsq_dim_c = std::get<0>(lsq_dim);
  static constexpr int lsq_dim_unk = std::get<1>(lsq_dim);

  // Parameter values.
  int i_startblk = 0;
  int i_endblk = nblks_c; // Test blocks [0 .. nblks_c-1]
  int i_startidx_in = 0;
  int i_endidx_in = nproma; // Full range: 0 .. nproma-1
  int slev = 0;
  int elev = nlev; // Full vertical range (0 .. nlev-1)
  int patch_id = 0;
  bool lacc = false;          // Not using ACC-specific behavior.
  bool acc_async = false;     // No asynchronous execution.
  bool l_consv = true;        // With conservative correction.
  bool l_limited_area = true; // Limited area setup

  std::vector<ValueType> p_cc;
  std::vector<int> cell_neighbor_idx;
  std::vector<int> cell_neighbor_blk;
  std::vector<ValueType> lsq_qtmat_c;
  std::vector<ValueType> lsq_rmat_rdiag_c;
  std::vector<ValueType> lsq_rmat_utri_c;
  std::vector<ValueType> lsq_moments;
  std::vector<ValueType> lsq_pseudoinv;
  std::vector<ValueType> p_coeff;

  HorizontalReconTest() {
    p_cc.resize(dim_combine(nproma, nlev, nblks_c));
    cell_neighbor_idx.resize(dim_combine(nproma, nblks_c, lsq_dim_c));
    cell_neighbor_blk.resize(dim_combine(nproma, nblks_c, lsq_dim_c));
    lsq_qtmat_c.resize(dim_combine(nproma, lsq_dim_unk, lsq_dim_c, nblks_c));
    lsq_rmat_rdiag_c.resize(dim_combine(nproma, lsq_dim_unk, nblks_c));
    lsq_rmat_utri_c.resize(dim_combine(
        nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c));
    lsq_moments.resize(dim_combine(nproma, nblks_c, lsq_dim_unk));
    lsq_pseudoinv.resize(dim_combine(nproma, lsq_dim_unk, lsq_dim_c, nblks_c));
    p_coeff.resize(dim_combine(lsq_dim_unk + 1, nproma, nlev, nblks_c));
  }
};

/// Test class for the horizontal tests. The reconstruction method is specified
/// to linear.
template <typename ValueType>
class HorizontalReconLinearTest
    : public HorizontalReconTest<ValueType, static_cast<int>(
                                                ReconstructionMethod::linear)> {
};

/// Test class for the horizontal tests. The reconstruction method is specified
/// to quadratic.
template <typename ValueType>
class HorizontalReconQuadraticTest
    : public HorizontalReconTest<
          ValueType, static_cast<int>(ReconstructionMethod::quadratic)> {};

/// Test class for the horizontal tests. The reconstruction method is specified
/// to cubic.
template <typename ValueType>
class HorizontalReconCubicTest
    : public HorizontalReconTest<ValueType, static_cast<int>(
                                                ReconstructionMethod::cubic)> {
};

/// ValueTypes which the divrot tests should run with
typedef ::testing::Types<float, double> ValueTypes;

TYPED_TEST_SUITE(HorizontalReconLinearTest, ValueTypes);

TYPED_TEST(HorizontalReconLinearTest, TestLsqCell) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 1)] = i;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 2)] = i;
    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
      this->lsq_qtmat_c[qtmat_at(i, 0, j, 0)] = 1.0;
      this->lsq_qtmat_c[qtmat_at(i, 1, j, 0)] = 0.5;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, 0, 0)] = 2.0;
    this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, 1, 0)] = 2.0;
    this->lsq_rmat_utri_c[rmat_utri_at(i, 0, 0)] = 0.1;

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
  }

  // Test function
  recon_lsq_cell_l<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->l_consv, this->lacc, this->acc_async,
      this->nblks_c, this->nlev, this->lsq_dim_unk, this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      0.34, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      1.8, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      1.0, 1e-6);
}

TYPED_TEST(HorizontalReconLinearTest, TestLsqCellRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 3.0);

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
      this->lsq_qtmat_c[qtmat_at(i, 0, j, 0)] = real_distrib(gen);
      this->lsq_qtmat_c[qtmat_at(i, 1, j, 0)] = real_distrib(gen);
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }

    this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, 0, 0)] = real_distrib(gen);
    this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, 1, 0)] = real_distrib(gen);
    this->lsq_rmat_utri_c[rmat_utri_at(i, 0, 0)] = real_distrib(gen);

    this->lsq_moments[moments_at(i, 0, 0)] = real_distrib(gen);
    this->lsq_moments[moments_at(i, 0, 1)] = real_distrib(gen);
  }

  // Test function
  recon_lsq_cell_l<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->l_consv, this->lacc, this->acc_async,
      this->nblks_c, this->nlev, this->lsq_dim_unk, this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> z_qt_times_d(lsq_dim_unk);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        z_qt_times_d[0] = 0.0;
        z_qt_times_d[1] = 0.0;
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_qt_times_d[0] += this->lsq_qtmat_c[qtmat_at(jc, 0, i, jb)] * z_d[i];
          z_qt_times_d[1] += this->lsq_qtmat_c[qtmat_at(jc, 1, i, jb)] * z_d[i];
        }
        p_result[at<lsq_dim_unk + 1, nproma>(2, jc)] =
            this->lsq_rmat_rdiag_c[rmat_rdiag_at(jc, 1, jb)] * z_qt_times_d[1];
        p_result[at<lsq_dim_unk + 1, nproma>(1, jc)] =
            this->lsq_rmat_rdiag_c[rmat_rdiag_at(jc, 0, jb)] *
            (z_qt_times_d[0] -
             this->lsq_rmat_utri_c[rmat_utri_at(jc, 0, jb)] *
                 p_result[at<lsq_dim_unk + 1, nproma>(2, jc)]);
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            this->p_cc[p_cc_at(jc, jk, jb)] -
            p_result[at<lsq_dim_unk + 1, nproma>(1, jc)] *
                this->lsq_moments[moments_at(jc, jb, 0)] -
            p_result[at<lsq_dim_unk + 1, nproma>(2, jc)] *
                this->lsq_moments[moments_at(jc, jb, 1)];
      }
    }
  }

  // Check result
  for (int i = 0; i < lsq_dim_unk + 1; ++i) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(i, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(i, jc))], 1e-5)
          << "For loop result fails for i = " << i << ", jc = " << jc;
    }
  }
}

TYPED_TEST(HorizontalReconLinearTest, TestLsqCellSVD) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 1)] = i;
    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 2)] = i;
    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
      this->lsq_pseudoinv[pseudoinv_at(i, 0, j, 0)] = 1.0;
      this->lsq_pseudoinv[pseudoinv_at(i, 1, j, 0)] = 0.5;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
  }

  // Test function
  recon_lsq_cell_l_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->l_consv, this->lacc, this->acc_async,
      this->nblks_c, this->nlev, this->lsq_dim_unk, this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      0.65, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      1.0, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      0.5, 1e-6);
}

TYPED_TEST(HorizontalReconLinearTest, TestLsqCellSVDRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 3.0);

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
      this->lsq_pseudoinv[pseudoinv_at(i, 0, j, 0)] = real_distrib(gen);
      this->lsq_pseudoinv[pseudoinv_at(i, 1, j, 0)] = real_distrib(gen);
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }

    this->lsq_moments[moments_at(i, 0, 0)] = real_distrib(gen);
    this->lsq_moments[moments_at(i, 0, 1)] = real_distrib(gen);
  }

  // Test function
  recon_lsq_cell_l_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->l_consv, this->lacc, this->acc_async,
      this->nblks_c, this->nlev, this->lsq_dim_unk, this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        p_result[at<lsq_dim_unk + 1, nproma>(2, jc)] =
            this->lsq_pseudoinv[pseudoinv_at(jc, 1, 0, jb)] * z_d[0] +
            this->lsq_pseudoinv[pseudoinv_at(jc, 1, 1, jb)] * z_d[1] +
            this->lsq_pseudoinv[pseudoinv_at(jc, 1, 2, jb)] * z_d[2];
        p_result[at<lsq_dim_unk + 1, nproma>(1, jc)] =
            this->lsq_pseudoinv[pseudoinv_at(jc, 0, 0, jb)] * z_d[0] +
            this->lsq_pseudoinv[pseudoinv_at(jc, 0, 1, jb)] * z_d[1] +
            this->lsq_pseudoinv[pseudoinv_at(jc, 0, 2, jb)] * z_d[2];
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
                this->p_cc[p_cc_at(jc, jk, jb)] -
                p_result[at<lsq_dim_unk + 1, nproma>(1, jc)] *
                    this->lsq_moments[moments_at(jc, jb, 0)] -
                p_result[at<lsq_dim_unk + 1, nproma>(2, jc)] *
                    this->lsq_moments[moments_at(jc, jb, 1)];
      }
    }
  }

  // Check result
  for (int i = 0; i < lsq_dim_unk + 1; ++i) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(i, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(i, jc))], 1e-5)
          << "For loop result fails for i = " << i << ", jc = " << jc;
    }
  }
}

TYPED_TEST_SUITE(HorizontalReconQuadraticTest, ValueTypes);

TYPED_TEST(HorizontalReconQuadraticTest, TestLsqCell) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_blk[cell_neighbor_at(i, 0, 0)] = 0;
    for (int j = 1; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = i;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->lsq_qtmat_c[qtmat_at(i, 0, j, 0)] = 1.0;
      this->lsq_qtmat_c[qtmat_at(i, 1, j, 0)] = 0.5;
      this->lsq_qtmat_c[qtmat_at(i, 2, j, 0)] = 0.2;
      this->lsq_qtmat_c[qtmat_at(i, 3, j, 0)] = 0.7;
      this->lsq_qtmat_c[qtmat_at(i, 4, j, 0)] = 1.3;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, j, 0)] = 2.0;
    }

    for (int j = 0; j < (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2; ++j) {
      this->lsq_rmat_utri_c[rmat_utri_at(i, j, 0)] = 1.0;
    }

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
    this->lsq_moments[moments_at(i, 0, 2)] = 0.4;
    this->lsq_moments[moments_at(i, 0, 3)] = 0.5;
    this->lsq_moments[moments_at(i, 0, 4)] = 0.6;
  }

  // Test function
  recon_lsq_cell_q<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      0.24, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      3.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      -2.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(3, 0, 0, 0))],
      2.8, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(4, 0, 0, 0))],
      -3.8, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(5, 0, 0, 0))],
      2.6, 1e-6);
}

TYPED_TEST(HorizontalReconQuadraticTest, TestLsqCellRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 1.0);

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      for (int k = 0; k < lsq_dim_c; ++k) {
        this->lsq_qtmat_c[qtmat_at(i, j, k, 0)] = real_distrib(gen);
      }
      this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, j, 0)] = real_distrib(gen);
      this->lsq_moments[moments_at(i, 0, j)] = real_distrib(gen);
    }
    for (int j = 0; j < (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2; ++j) {
      this->lsq_rmat_utri_c[rmat_utri_at(i, j, 0)] = real_distrib(gen);
    }

    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }
  }

  // Test function
  recon_lsq_cell_q<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> z_qt_times_d(lsq_dim_unk);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        for (int j = 0; j < lsq_dim_unk; ++j) {
          z_qt_times_d[j] = 0.0;
          for (int i = 0; i < lsq_dim_c; ++i) {
            z_qt_times_d[j] +=
                this->lsq_qtmat_c[qtmat_at(jc, j, i, jb)] * z_d[i];
          }
        }
        int utri_id = 0;
        for (int j = lsq_dim_unk; j > 0; --j) {
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] = z_qt_times_d[j - 1];
          for (int k = j + 1; k <= lsq_dim_unk; ++k) {
            p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] -=
                this->lsq_rmat_utri_c[rmat_utri_at(jc, utri_id++, jb)] *
                p_result[at<lsq_dim_unk + 1, nproma>(k, jc)];
          }
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] *=
              this->lsq_rmat_rdiag_c[rmat_rdiag_at(jc, j - 1, jb)];
        }
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            this->p_cc[p_cc_at(jc, jk, jb)];
        for (int j = 0; j < lsq_dim_unk; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] -=
              p_result[at<lsq_dim_unk + 1, nproma>(j + 1, jc)] *
              this->lsq_moments[moments_at(jc, jb, j)];
        }
      }
    }
  }

  // Check result
  for (int i = 0; i < lsq_dim_unk + 1; ++i) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(i, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(i, jc))], 1e-5)
          << "For loop result fails for i = " << i << ", jc = " << jc;
    }
  }
}

TYPED_TEST(HorizontalReconQuadraticTest, TestLsqCellSVD) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_blk[cell_neighbor_at(i, 0, 0)] = 0;
    for (int j = 1; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = i;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->lsq_pseudoinv[pseudoinv_at(i, 0, j, 0)] = 1.0;
      this->lsq_pseudoinv[pseudoinv_at(i, 1, j, 0)] = 0.5;
      this->lsq_pseudoinv[pseudoinv_at(i, 2, j, 0)] = 0.2;
      this->lsq_pseudoinv[pseudoinv_at(i, 3, j, 0)] = 0.7;
      this->lsq_pseudoinv[pseudoinv_at(i, 4, j, 0)] = 1.3;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
    this->lsq_moments[moments_at(i, 0, 2)] = 0.4;
    this->lsq_moments[moments_at(i, 0, 3)] = 0.5;
    this->lsq_moments[moments_at(i, 0, 4)] = 0.6;
  }

  // Test function
  recon_lsq_cell_q_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      -0.56, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      1.0, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      0.5, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(3, 0, 0, 0))],
      0.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(4, 0, 0, 0))],
      0.7, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(5, 0, 0, 0))],
      1.3, 1e-6);
}

TYPED_TEST(HorizontalReconQuadraticTest, TestLsqCellSVDRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 1.0);

  // Initialization is done only for iblk = 0 and ilev = 0
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      for (int k = 0; k < lsq_dim_c; ++k) {
        this->lsq_pseudoinv[pseudoinv_at(i, j, k, 0)] = real_distrib(gen);
      }
      this->lsq_moments[moments_at(i, 0, j)] = real_distrib(gen);
    }

    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }
  }

  // Test function
  recon_lsq_cell_q_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> z_qt_times_d(lsq_dim_unk);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        for (int j = 1; j < lsq_dim_unk + 1; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] = 0.0;
          for (int i = 0; i < lsq_dim_c; ++i) {
            p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] +=
                this->lsq_pseudoinv[pseudoinv_at(jc, j - 1, i, jb)] * z_d[i];
          }
        }
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            this->p_cc[p_cc_at(jc, jk, jb)];
        for (int j = 0; j < lsq_dim_unk; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] -=
              p_result[at<lsq_dim_unk + 1, nproma>(j + 1, jc)] *
              this->lsq_moments[moments_at(jc, jb, j)];
        }
      }
    }
  }

  // Check result
  for (int j = 0; j < lsq_dim_unk + 1; ++j) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(j, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(j, jc))], 1e-5)
          << "For loop result fails for j = " << j << ", jc = " << jc;
    }
  }
}

TYPED_TEST_SUITE(HorizontalReconCubicTest, ValueTypes);

TYPED_TEST(HorizontalReconCubicTest, TestLsqCell) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_blk[cell_neighbor_at(i, 0, 0)] = 0;
    for (int j = 1; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = i;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->lsq_qtmat_c[qtmat_at(i, 0, j, 0)] = 1.0;
      this->lsq_qtmat_c[qtmat_at(i, 1, j, 0)] = 0.9;
      this->lsq_qtmat_c[qtmat_at(i, 2, j, 0)] = 0.8;
      this->lsq_qtmat_c[qtmat_at(i, 3, j, 0)] = 0.7;
      this->lsq_qtmat_c[qtmat_at(i, 4, j, 0)] = 0.6;
      this->lsq_qtmat_c[qtmat_at(i, 5, j, 0)] = 0.5;
      this->lsq_qtmat_c[qtmat_at(i, 6, j, 0)] = 0.4;
      this->lsq_qtmat_c[qtmat_at(i, 7, j, 0)] = 0.3;
      this->lsq_qtmat_c[qtmat_at(i, 8, j, 0)] = 0.2;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, j, 0)] = 2.0;
    }

    for (int j = 0; j < (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2; ++j) {
      this->lsq_rmat_utri_c[rmat_utri_at(i, j, 0)] = 1.0;
    }

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
    this->lsq_moments[moments_at(i, 0, 2)] = 0.4;
    this->lsq_moments[moments_at(i, 0, 3)] = 0.5;
    this->lsq_moments[moments_at(i, 0, 4)] = 0.6;
    this->lsq_moments[moments_at(i, 0, 5)] = 0.7;
    this->lsq_moments[moments_at(i, 0, 6)] = 0.8;
    this->lsq_moments[moments_at(i, 0, 7)] = 0.9;
    this->lsq_moments[moments_at(i, 0, 8)] = 1.0;
  }

  // Test function
  recon_lsq_cell_c<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      0.28, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      0.4, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      -0.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(3, 0, 0, 0))],
      0.4, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(4, 0, 0, 0))],
      -0.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(5, 0, 0, 0))],
      0.4, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(6, 0, 0, 0))],
      -0.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(7, 0, 0, 0))],
      0.4, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(8, 0, 0, 0))],
      -0.2, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(9, 0, 0, 0))],
      0.4, 1e-6);
}

TYPED_TEST(HorizontalReconCubicTest, TestLsqCellRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &qtmat_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &rmat_rdiag_at = at<nproma, lsq_dim_unk, nblks_c>;
  const auto &rmat_utri_at =
      at<nproma, (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 1.0);

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      for (int k = 0; k < lsq_dim_c; ++k) {
        this->lsq_qtmat_c[qtmat_at(i, j, k, 0)] = real_distrib(gen);
      }
      this->lsq_rmat_rdiag_c[rmat_rdiag_at(i, j, 0)] = real_distrib(gen);
      this->lsq_moments[moments_at(i, 0, j)] = real_distrib(gen);
    }
    for (int j = 0; j < (lsq_dim_unk * lsq_dim_unk - lsq_dim_unk) / 2; ++j) {
      this->lsq_rmat_utri_c[rmat_utri_at(i, j, 0)] = real_distrib(gen);
    }

    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }
  }

  // Test function
  recon_lsq_cell_c<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_qtmat_c.data(),
      this->lsq_rmat_rdiag_c.data(), this->lsq_rmat_utri_c.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> z_qt_times_d(lsq_dim_unk);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        for (int j = 0; j < lsq_dim_unk; ++j) {
          z_qt_times_d[j] = 0.0;
          for (int i = 0; i < lsq_dim_c; ++i) {
            z_qt_times_d[j] +=
                this->lsq_qtmat_c[qtmat_at(jc, j, i, jb)] * z_d[i];
          }
        }
        int utri_id = 0;
        for (int j = lsq_dim_unk; j > 0; --j) {
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] = z_qt_times_d[j - 1];
          for (int k = j + 1; k <= lsq_dim_unk; ++k) {
            p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] -=
                this->lsq_rmat_utri_c[rmat_utri_at(jc, utri_id++, jb)] *
                p_result[at<lsq_dim_unk + 1, nproma>(k, jc)];
          }
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] *=
              this->lsq_rmat_rdiag_c[rmat_rdiag_at(jc, j - 1, jb)];
        }
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            this->p_cc[p_cc_at(jc, jk, jb)];
        for (int j = 0; j < lsq_dim_unk; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] -=
              p_result[at<lsq_dim_unk + 1, nproma>(j + 1, jc)] *
              this->lsq_moments[moments_at(jc, jb, j)];
        }
      }
    }
  }

  // Check result
  for (int i = 0; i < lsq_dim_unk + 1; ++i) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(i, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(i, jc))], 1e-5)
          << "For loop result fails for i = " << i << ", jc = " << jc;
    }
  }
}

TYPED_TEST(HorizontalReconCubicTest, TestLsqCellSVD) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = (i + 1);

    this->cell_neighbor_idx[cell_neighbor_at(i, 0, 0)] = (i + 1) % nproma;
    this->cell_neighbor_blk[cell_neighbor_at(i, 0, 0)] = 0;
    for (int j = 1; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = i;
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->lsq_pseudoinv[pseudoinv_at(i, 0, j, 0)] = 1.0;
      this->lsq_pseudoinv[pseudoinv_at(i, 1, j, 0)] = 0.9;
      this->lsq_pseudoinv[pseudoinv_at(i, 2, j, 0)] = 0.8;
      this->lsq_pseudoinv[pseudoinv_at(i, 3, j, 0)] = 0.7;
      this->lsq_pseudoinv[pseudoinv_at(i, 4, j, 0)] = 0.6;
      this->lsq_pseudoinv[pseudoinv_at(i, 5, j, 0)] = 0.5;
      this->lsq_pseudoinv[pseudoinv_at(i, 6, j, 0)] = 0.4;
      this->lsq_pseudoinv[pseudoinv_at(i, 7, j, 0)] = 0.3;
      this->lsq_pseudoinv[pseudoinv_at(i, 8, j, 0)] = 0.2;
    }
    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = 0.0;
    }

    this->lsq_moments[moments_at(i, 0, 0)] = 0.2;
    this->lsq_moments[moments_at(i, 0, 1)] = 0.3;
    this->lsq_moments[moments_at(i, 0, 2)] = 0.4;
    this->lsq_moments[moments_at(i, 0, 3)] = 0.5;
    this->lsq_moments[moments_at(i, 0, 4)] = 0.6;
    this->lsq_moments[moments_at(i, 0, 5)] = 0.7;
    this->lsq_moments[moments_at(i, 0, 6)] = 0.8;
    this->lsq_moments[moments_at(i, 0, 7)] = 0.9;
    this->lsq_moments[moments_at(i, 0, 8)] = 1.0;
  }

  // Test function
  recon_lsq_cell_c_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Check result
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(0, 0, 0, 0))],
      -1.64, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(1, 0, 0, 0))],
      1.0, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(2, 0, 0, 0))],
      0.9, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(3, 0, 0, 0))],
      0.8, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(4, 0, 0, 0))],
      0.7, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(5, 0, 0, 0))],
      0.6, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(6, 0, 0, 0))],
      0.5, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(7, 0, 0, 0))],
      0.4, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(8, 0, 0, 0))],
      0.3, 1e-6);
  EXPECT_NEAR(
      this->p_coeff[(at<lsq_dim_unk + 1, nproma, nlev, nblks_c>(9, 0, 0, 0))],
      0.2, 1e-6);
}

TYPED_TEST(HorizontalReconCubicTest, TestLsqCellSVDRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_c = this->nblks_c;
  constexpr int lsq_dim_c = this->lsq_dim_c;
  constexpr int lsq_dim_unk = this->lsq_dim_unk;

  const auto &p_cc_at = at<nproma, nlev, nblks_c>;
  const auto &cell_neighbor_at = at<nproma, nblks_c, lsq_dim_c>;
  const auto &pseudoinv_at = at<nproma, lsq_dim_unk, lsq_dim_c, nblks_c>;
  const auto &p_coeff_at = at<lsq_dim_unk + 1, nproma, nlev, nblks_c>;
  const auto &moments_at = at<nproma, nblks_c, lsq_dim_unk>;

  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(0.0, 1.0);

  // Initialization
  for (int i = 0; i < nproma; ++i) {
    this->p_cc[p_cc_at(i, 0, 0)] = real_distrib(gen);

    for (int j = 0; j < lsq_dim_c; ++j) {
      this->cell_neighbor_idx[cell_neighbor_at(i, 0, j)] = int_distrib(gen);
      this->cell_neighbor_blk[cell_neighbor_at(i, 0, j)] = 0;
    }

    for (int j = 0; j < lsq_dim_unk; ++j) {
      for (int k = 0; k < lsq_dim_c; ++k) {
        this->lsq_pseudoinv[pseudoinv_at(i, j, k, 0)] = real_distrib(gen);
      }
      this->lsq_moments[moments_at(i, 0, j)] = real_distrib(gen);
    }

    for (int j = 0; j < lsq_dim_unk + 1; ++j) {
      this->p_coeff[p_coeff_at(j, i, 0, 0)] = real_distrib(gen);
    }
  }

  // Test function
  recon_lsq_cell_c_svd<TypeParam>(
      this->p_cc.data(), this->cell_neighbor_idx.data(),
      this->cell_neighbor_blk.data(), this->lsq_pseudoinv.data(),
      this->lsq_moments.data(), this->p_coeff.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->patch_id, this->l_limited_area,
      this->lacc, this->nblks_c, this->nlev, this->lsq_dim_unk,
      this->lsq_dim_c);

  // Compute reference result
  std::vector<TypeParam> z_d(lsq_dim_c);
  std::vector<TypeParam> z_qt_times_d(lsq_dim_unk);
  std::vector<TypeParam> p_result((lsq_dim_unk + 1) * nproma);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_c_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);
    for (int jk = this->slev; jk < this->elev; ++jk) {
      for (int jc = i_startidx; jc < i_endidx; ++jc) {
        for (int i = 0; i < lsq_dim_c; ++i) {
          z_d[i] = this->p_cc[p_cc_at(
                       this->cell_neighbor_idx[cell_neighbor_at(jc, jb, i)], jk,
                       this->cell_neighbor_blk[cell_neighbor_at(jc, jb, i)])] -
                   this->p_cc[p_cc_at(jc, jk, jb)];
        }
        for (int j = 1; j < lsq_dim_unk + 1; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] = 0.0;
          for (int i = 0; i < lsq_dim_c; ++i) {
            p_result[at<lsq_dim_unk + 1, nproma>(j, jc)] +=
                this->lsq_pseudoinv[pseudoinv_at(jc, j - 1, i, jb)] * z_d[i];
          }
        }
        p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] =
            this->p_cc[p_cc_at(jc, jk, jb)];
        for (int j = 0; j < lsq_dim_unk; ++j) {
          p_result[at<lsq_dim_unk + 1, nproma>(0, jc)] -=
              p_result[at<lsq_dim_unk + 1, nproma>(j + 1, jc)] *
              this->lsq_moments[moments_at(jc, jb, j)];
        }
      }
    }
  }
  // Check result
  for (int i = 0; i < lsq_dim_unk + 1; ++i) {
    for (int jc = 0; jc < nproma; ++jc) {
      EXPECT_NEAR(this->p_coeff[(p_coeff_at(i, jc, 0, 0))],
                  p_result[(at<lsq_dim_unk + 1, nproma>(i, jc))], 1e-5)
          << "For loop result fails for i = " << i << ", jc = " << jc;
    }
  }
}
