// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include <iostream>
#include <random>
#include <vector>

#include <Kokkos_Core.hpp>
#include <gtest/gtest.h>
#include <dim_helper.hpp>
#include <horizontal/mo_lib_divrot.hpp>
#include <support/mo_lib_loopindices.hpp>

/// Test class for the horizontal rotation tests. Templated for the ValueType.
template <typename ValueType>
class HorizontalRotVertexTest : public ::testing::Test {
protected:
  static constexpr int nproma = 3;  // inner loop length
  static constexpr int nlev = 2;    // number of vertical levels
  static constexpr int nblks_e = 1; // number of edge blocks
  static constexpr int nblks_v = 1; // number of vertex blocks
  static constexpr int dim4d = 2;   // 4th dimension size

  int i_startblk = 0;
  int i_endblk = nblks_v; // Test blocks [0 .. nblks_v-1]
  int i_startidx_in = 0;
  int i_endidx_in = nproma; // Full range: 0 .. nproma-1
  std::vector<int> slev;
  std::vector<int> elev;
  bool lacc = false;      // Not using ACC-specific behavior.
  bool acc_async = false; // Not using ACC-specific behavior.

  std::vector<ValueType> vec_e;
  std::vector<int> vert_edge_idx;
  std::vector<int> vert_edge_blk;
  std::vector<ValueType> geofac_rot;
  std::vector<ValueType> rot_vec;
  std::vector<ValueType> f4din;
  std::vector<ValueType> f4dout;

  HorizontalRotVertexTest() {
    slev.resize(dim4d, 0);
    elev.resize(dim4d, nlev); // Full vertical range (0 .. nlev-1)

    vec_e.resize(dim_combine(nproma, nlev, nblks_e));
    vert_edge_idx.resize(dim_combine(nproma, nblks_v, 6));
    vert_edge_blk.resize(dim_combine(nproma, nblks_v, 6));
    geofac_rot.resize(dim_combine(nproma, 6, nblks_v));
    rot_vec.resize(dim_combine(nproma, nlev, nblks_v));
    f4din.resize(dim_combine(nproma, nlev, nblks_e, dim4d));
    f4dout.resize(dim_combine(nproma, nlev, nblks_v, dim4d));
  }
};

/// ValueTypes which the divrot tests should run with
typedef ::testing::Types<float, double> ValueTypes;

TYPED_TEST_SUITE(HorizontalRotVertexTest, ValueTypes);

TYPED_TEST(HorizontalRotVertexTest, TestRotVertexAtmosSpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_e = this->nblks_e;
  constexpr int nblks_v = this->nblks_v;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &vert_edge_at = at<nproma, nblks_v, 6>;
  const auto &geofac_rot_at = at<nproma, 6, nblks_v>;
  const auto &rot_vec_at = at<nproma, nlev, nblks_v>;

  // Initialization with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
    }

    // Set edge indices to point to specific edges
    for (int j = 0; j < 6; ++j) {
      this->vert_edge_idx[vert_edge_at(i, 0, j)] = (i + j) % nproma;
      // All edges are in the same block for this test
      this->vert_edge_blk[vert_edge_at(i, 0, j)] = 0;
    }

    // Geometric factors for rotation
    this->geofac_rot[geofac_rot_at(i, 0, 0)] = 0.3;
    this->geofac_rot[geofac_rot_at(i, 1, 0)] = 0.2;
    this->geofac_rot[geofac_rot_at(i, 2, 0)] = 0.1;
    this->geofac_rot[geofac_rot_at(i, 3, 0)] = 0.2;
    this->geofac_rot[geofac_rot_at(i, 4, 0)] = 0.1;
    this->geofac_rot[geofac_rot_at(i, 5, 0)] = 0.1;

    // Initialize rot_vec to zero
    for (int k = 0; k < nlev; ++k) {
      this->rot_vec[rot_vec_at(i, k, 0)] = 0.0;
    }
  }

  // Call the rot_vertex_atmos function
  rot_vertex_atmos<TypeParam>(
      this->vec_e.data(), this->vert_edge_idx.data(),
      this->vert_edge_blk.data(), this->geofac_rot.data(), this->rot_vec.data(),
      this->i_startblk, this->i_endblk, this->i_startidx_in, this->i_endidx_in,
      this->slev[0], this->elev[0], this->nproma, this->lacc, this->nlev,
      this->nblks_e, this->nblks_v);

  // Expected values based on the initialization pattern
  EXPECT_NEAR(this->rot_vec[rot_vec_at(0, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(0, 1, 0)], 3.4, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(1, 0, 0)], 2.1, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(1, 1, 0)], 4.2, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(2, 0, 0)], 2.2, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(2, 1, 0)], 4.4, 1e-6);
}

TYPED_TEST(HorizontalRotVertexTest, TestRotVertexAtmosRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_e = this->nblks_e;
  constexpr int nblks_v = this->nblks_v;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &vert_edge_at = at<nproma, nblks_v, 6>;
  const auto &geofac_rot_at = at<nproma, 6, nblks_v>;
  const auto &rot_vec_at = at<nproma, nlev, nblks_v>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialization with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
    }

    // Set random edge indices
    for (int j = 0; j < 6; ++j) {
      this->vert_edge_idx[vert_edge_at(i, 0, j)] = int_distrib(gen);
      this->vert_edge_blk[vert_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 6; ++j) {
      this->geofac_rot[geofac_rot_at(i, j, 0)] = real_distrib(gen);
    }

    // Initialize rot_vec to random values
    for (int k = 0; k < nlev; ++k) {
      this->rot_vec[rot_vec_at(i, k, 0)] = real_distrib(gen);
    }
  }

  // Call the rot_vertex_atmos function
  rot_vertex_atmos<TypeParam>(
      this->vec_e.data(), this->vert_edge_idx.data(),
      this->vert_edge_blk.data(), this->geofac_rot.data(), this->rot_vec.data(),
      this->i_startblk, this->i_endblk, this->i_startidx_in, this->i_endidx_in,
      this->slev[0], this->elev[0], this->nproma, this->lacc, this->nlev,
      this->nblks_e, this->nblks_v);

  // Calculate reference values separately and verify results
  std::vector<TypeParam> ref_rot_vec(nproma * nlev * nblks_v, 0.0);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_v_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jv = i_startidx; jv < i_endidx; ++jv) {
        ref_rot_vec[rot_vec_at(jv, jk, jb)] =
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 0)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 0)])] *
                this->geofac_rot[geofac_rot_at(jv, 0, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 1)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 1)])] *
                this->geofac_rot[geofac_rot_at(jv, 1, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 2)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 2)])] *
                this->geofac_rot[geofac_rot_at(jv, 2, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 3)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 3)])] *
                this->geofac_rot[geofac_rot_at(jv, 3, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 4)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 4)])] *
                this->geofac_rot[geofac_rot_at(jv, 4, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 5)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 5)])] *
                this->geofac_rot[geofac_rot_at(jv, 5, jb)];
      }
    }
  }

  // Verify results
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->rot_vec[rot_vec_at(i, k, 0)],
                  ref_rot_vec[rot_vec_at(i, k, 0)], 1e-5)
          << "Results differ at i=" << i << ", k=" << k;
    }
  }
}

TYPED_TEST_SUITE(HorizontalRotVertexTest, ValueTypes);

TYPED_TEST(HorizontalRotVertexTest, TestRotVertexRISpecific) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_e = this->nblks_e;
  constexpr int nblks_v = this->nblks_v;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &vert_edge_at = at<nproma, nblks_v, 6>;
  const auto &geofac_rot_at = at<nproma, 6, nblks_v>;
  const auto &rot_vec_at = at<nproma, nlev, nblks_v>;

  // Initialization with specific values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = (i + 1) * (k + 1); // Simple pattern
    }

    // Set edge indices to point to specific edges
    for (int j = 0; j < 6; ++j) {
      this->vert_edge_idx[vert_edge_at(i, 0, j)] = (i + j) % nproma;
      // All edges are in the same block for this test
      this->vert_edge_blk[vert_edge_at(i, 0, j)] = 0;
    }

    // Geometric factors for rotation
    this->geofac_rot[geofac_rot_at(i, 0, 0)] = 0.3;
    this->geofac_rot[geofac_rot_at(i, 1, 0)] = 0.2;
    this->geofac_rot[geofac_rot_at(i, 2, 0)] = 0.1;
    this->geofac_rot[geofac_rot_at(i, 3, 0)] = 0.2;
    this->geofac_rot[geofac_rot_at(i, 4, 0)] = 0.1;
    this->geofac_rot[geofac_rot_at(i, 5, 0)] = 0.1;

    // Initialize rot_vec to zero
    for (int k = 0; k < nlev; ++k) {
      this->rot_vec[rot_vec_at(i, k, 0)] = 0.0;
    }
  }

  // Call the rot_vertex_ri function
  rot_vertex_ri<TypeParam>(
      this->vec_e.data(), this->vert_edge_idx.data(),
      this->vert_edge_blk.data(), this->geofac_rot.data(), this->rot_vec.data(),
      this->i_startblk, this->i_endblk, this->i_startidx_in, this->i_endidx_in,
      this->slev[0], this->elev[0], this->nproma, this->lacc, this->acc_async,
      this->nlev, this->nblks_e, this->nblks_v);

  // Expected values based on the initialization pattern
  EXPECT_NEAR(this->rot_vec[rot_vec_at(0, 0, 0)], 1.7, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(0, 1, 0)], 3.4, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(1, 0, 0)], 2.1, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(1, 1, 0)], 4.2, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(2, 0, 0)], 2.2, 1e-6);
  EXPECT_NEAR(this->rot_vec[rot_vec_at(2, 1, 0)], 4.4, 1e-6);
}

TYPED_TEST(HorizontalRotVertexTest, TestRotVertexRIRandom) {
  constexpr int nproma = this->nproma;
  constexpr int nlev = this->nlev;
  constexpr int nblks_e = this->nblks_e;
  constexpr int nblks_v = this->nblks_v;

  const auto &vec_e_at = at<nproma, nlev, nblks_e>;
  const auto &vert_edge_at = at<nproma, nblks_v, 6>;
  const auto &geofac_rot_at = at<nproma, 6, nblks_v>;
  const auto &rot_vec_at = at<nproma, nlev, nblks_v>;

  // Set up random number generators
  std::random_device rd;
  std::mt19937 gen(rd());
  std::uniform_int_distribution<int> int_distrib(0, nproma - 1);
  std::uniform_real_distribution<TypeParam> real_distrib(-10.0, 10.0);

  // Initialization with random values
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      this->vec_e[vec_e_at(i, k, 0)] = real_distrib(gen);
    }

    // Set random edge indices
    for (int j = 0; j < 6; ++j) {
      this->vert_edge_idx[vert_edge_at(i, 0, j)] = int_distrib(gen);
      this->vert_edge_blk[vert_edge_at(i, 0, j)] =
          0; // Keep in same block for simplicity
    }

    // Random geometric factors
    for (int j = 0; j < 6; ++j) {
      this->geofac_rot[geofac_rot_at(i, j, 0)] = real_distrib(gen);
    }

    // Initialize rot_vec to random values
    for (int k = 0; k < nlev; ++k) {
      this->rot_vec[rot_vec_at(i, k, 0)] = real_distrib(gen);
    }
  }

  // Call the rot_vertex_ri function
  rot_vertex_ri<TypeParam>(
      this->vec_e.data(), this->vert_edge_idx.data(),
      this->vert_edge_blk.data(), this->geofac_rot.data(), this->rot_vec.data(),
      this->i_startblk, this->i_endblk, this->i_startidx_in, this->i_endidx_in,
      this->slev[0], this->elev[0], this->nproma, this->lacc, this->acc_async,
      this->nlev, this->nblks_e, this->nblks_v);

  // Ensure computation is complete for both modes
  Kokkos::fence();

  // Calculate reference values separately and verify results
  std::vector<TypeParam> ref_rot_vec(nproma * nlev * nblks_v, 0.0);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    int i_startidx, i_endidx;
    get_indices_v_lib(this->i_startidx_in, this->i_endidx_in, nproma, jb,
                      this->i_startblk, this->i_endblk, i_startidx, i_endidx);

    for (int jk = this->slev[0]; jk < this->elev[0]; ++jk) {
      for (int jv = i_startidx; jv < i_endidx; ++jv) {
        ref_rot_vec[rot_vec_at(jv, jk, jb)] =
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 0)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 0)])] *
                this->geofac_rot[geofac_rot_at(jv, 0, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 1)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 1)])] *
                this->geofac_rot[geofac_rot_at(jv, 1, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 2)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 2)])] *
                this->geofac_rot[geofac_rot_at(jv, 2, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 3)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 3)])] *
                this->geofac_rot[geofac_rot_at(jv, 3, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 4)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 4)])] *
                this->geofac_rot[geofac_rot_at(jv, 4, jb)] +
            this->vec_e[vec_e_at(
                this->vert_edge_idx[vert_edge_at(jv, jb, 5)], jk,
                this->vert_edge_blk[vert_edge_at(jv, jb, 5)])] *
                this->geofac_rot[geofac_rot_at(jv, 5, jb)];
      }
    }
  }

  // Verify results
  for (int i = 0; i < nproma; ++i) {
    for (int k = 0; k < nlev; ++k) {
      EXPECT_NEAR(this->rot_vec[rot_vec_at(i, k, 0)],
                  ref_rot_vec[rot_vec_at(i, k, 0)], 1e-5)
          << "Results differ at i=" << i << ", k=" << k << ")";
    }
  }
}

