// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2025, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include <gtest/gtest.h>
#include <vector>
#include <algorithm>
#include "mo_math_utilities.hpp"

// Helper function to compute the 1D index for column-major storage.
inline int idx(int i, int j, int nrows) {
  return i + j * nrows;
}

// Test fixture for the TDMA solver tests.
class TDMASolverTestFixture : public ::testing::Test {
protected:
  const int n = 10;             // Matrix dimension.
  std::vector<double> a;        // Input matrix a.
  std::vector<double> b;        // Input matrix b.
  std::vector<double> c;        // Input matrix c.
  std::vector<double> d;        // Input matrix d.
  std::vector<double> x;        // Output matrix.

  TDMASolverTestFixture()
      : a(n * n), b(n * n), c(n * n), d(n * n), x(n * n, 0.0) {}

  // SetUp is run before each test.
  void SetUp() override {
    // Fill arrays in column-major order.
    for (int j = 0; j < n; j++) {
      for (int i = 0; i < n; i++) {
        double value = (i + 1) + (j + 1);
        a[idx(i, j, n)] = 1.0 * value;
        b[idx(i, j, n)] = 2.0 * value;
        c[idx(i, j, n)] = 1.0 * value;
        d[idx(i, j, n)] = 1.0 * value;
      }
    }
    // Clear the output vector.
    std::fill(x.begin(), x.end(), 0.0);
  }
};

TEST_F(TDMASolverTestFixture, FullTest) {
  // Call the solver over the full range:
  tdma_solver_vec<double>(a.data(), b.data(), c.data(), d.data(),
                         0, n, 0, n, n, n, x.data());

  // Compute the sum of all elements in the output matrix.
  double sum = 0.0;
  for (int j = 0; j < n; j++) {
    for (int i = 0; i < n; i++) {
      sum += x[idx(i, j, n)];
    }
  }

  // Expected reference sum
  double sum_ref = 27.2727272727272769;
  double tol = 1e-13;
  EXPECT_NEAR(sum, sum_ref, tol);
}

TEST_F(TDMASolverTestFixture, PartialTest) {
  // Call the solver for a partial region:
  // For C++: slev = 1, elev = n-1, startidx = 1, endidx = n-1.
  tdma_solver_vec<double>(a.data(), b.data(), c.data(), d.data(),
                         1, n - 1, 1, n - 1, n, n, x.data());

  // Compute the sum over a region
  double sum = 0.0;
  for (int j = 1; j < n - 1; j++) {
    for (int i = 1; i < n - 1; i++) {
      sum += x[idx(i, j, n)];
    }
  }

  double sum_ref = 17.7777777777777679;
  double tol = 1e-13;
  EXPECT_NEAR(sum, sum_ref, tol);
}
